@@ -3,6 +3,7 @@ import os
 import py_compile
 import shutil
 import stat
+import sys
 import tempfile
 import unittest
 
@@ -99,5 +100,25 @@ class PyCompileTests(unittest.TestCase):
         self.assertFalse(os.path.exists(
             importlib.util.cache_from_source(bad_coding)))
 
+    def test_double_dot_no_clobber(self):
+        # http://bugs.python.org/issue22966
+        # py_compile foo.bar.py -> __pycache__/foo.cpython-34.pyc
+        weird_path = os.path.join(self.directory, 'foo.bar.py')
+        cache_path = importlib.util.cache_from_source(weird_path)
+        pyc_path = weird_path + 'c'
+        head, tail = os.path.split(cache_path)
+        penultimate_tail = os.path.basename(head)
+        self.assertEqual(
+            os.path.join(penultimate_tail, tail),
+            os.path.join(
+                '__pycache__',
+                'foo.bar.{}.pyc'.format(sys.implementation.cache_tag)))
+        with open(weird_path, 'w') as file:
+            file.write('x = 123\n')
+        py_compile.compile(weird_path)
+        self.assertTrue(os.path.exists(cache_path))
+        self.assertFalse(os.path.exists(pyc_path))
+
+
 if __name__ == "__main__":
     unittest.main()
